from collections import Counter
from typing import List, Optional

from .constants import SUPPORTED_ITEM_TYPES
from .models import Result
from .utils import cleanup_string, default_checker, heap_sort


def binary_search(term: SUPPORTED_ITEM_TYPES, items: List[SUPPORTED_ITEM_TYPES]) -> Optional[Result]:
    """
    Performs a binary search on a list of integers to find a target value.

    Examples:
        >>> huntela.binary_search(term='a', items=['a', 'b', 'c'])
        {'confidence': 1, 'index': 0, 'value': 'a'}
        >>> huntela.binary_search(term=5, items=[1, 5, 10, 15])
        {'confidence': 1, 'index': 1, 'value': 5}

    Args:
        term: The target integer to search for.
        items: The list of integers to search through.

    Returns:
        The index of the target value in the list, or None if it is not found.
    """

    left, right = 0, len(items) - 1

    while left <= right:
        mid = (left + right) // 2
        if items[mid] == term:
            return Result(confidence=1, index=mid, value=term)
        elif items[mid] > term:
            right = mid - 1
        else:
            left = mid + 1

    return None


def simple_search(term: SUPPORTED_ITEM_TYPES, items: List[SUPPORTED_ITEM_TYPES]) -> List[Result]:
    """
    Searches a list of items for a given search term.

    Examples:
        >>> huntela.simple_search("app", ["app", "apple", "hello", "world"])
        [
            {{'confidence': 1, 'index': 0, 'value': 'app'}},
            {{'confidence': 0.6, 'index': 1, 'value': 'apple'}}
        ]

    Args:
        term (SUPPORTED_ITEM_TYPES): The search term to match against items in the list.
        items (List[SUPPORTED_ITEM_TYPES]): The list of items to search.

    Returns:
        {List[Result]}: A list of Result objects representing the search results.
    """

    results = []

    for index in range(len(items)):
        item = items[index]

        if type(item) is str:
            item = cleanup_string(item)

        match = default_checker(item, term)
        if match[0]:
            results.append(Result(index=index, value=item, confidence=match[1]))

    results.sort(key=lambda x: x['confidence'])

    return results


def search_for_least_frequent_items(size: int, items: List[SUPPORTED_ITEM_TYPES]):
    """
    Finds the k least frequent item(s) in a list.

    Args:
        size (int): The number of least frequent items to return.
        items (List[SUPPORTED_ITEM_TYPES]): A list of items.

    Returns:
        A list of the least frequent item(s).
    """

    results = []
    for item, (frequency, indices) in heap_sort(size, items, 'ASC'):
        results.append(Result(confidence=1, value=item, index=indices))
        
    return results


def search_for_most_frequent_items(size: int, items: List[SUPPORTED_ITEM_TYPES]):
    """
    Finds the k most frequent item(s) in a list.

    Args:
        size (int): The number of most frequent items to return.
        items (List[SUPPORTED_ITEM_TYPES]): A list of items.

    Returns:
        A list of the most frequent item(s).
    """

    results = []
    for item, (frequency, indices) in heap_sort(size, items, 'DESC'):
        results.append(Result(confidence=1, value=item, index=indices))
        
    return results


def search_csv_file(filename: str, column: str, value: str):
    raise NotImplementedError



import os
from setuptools import setup, find_packages

current_dir = os.path.dirname(os.path.abspath(__file__))
parent_dir = os.path.abspath(os.path.join(current_dir, os.pardir))
readme_path = os.path.join(parent_dir, "README.md")

with open(readme_path, "rt") as f:
    readme = f.read()

VERSION = '0.0.13' 
DESCRIPTION = "Find what you're looking for in a flash with Huntela."
LONG_DESCRIPTION = readme

# Setting up
setup(
    # the name must match the folder name 'verysimplemodule'
    name="huntela", 
    version=VERSION,
    author="Tomisin Abiodun",
    author_email="decave.12357@gmail.com",
    description=DESCRIPTION,
    url='https://github.com/CodeLexis/huntela',
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=[], # add any additional packages that 
    # needs to be installed along with your package. Eg: 'caer'
    python_requires=">=3.9, <4",
    keywords=['python', 'search', 'query'],
    classifiers= [
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Topic :: Software Development :: Libraries"
    ]
)



"""Provide several sample math calculations.

This module allows the user to make mathematical calculations.

The module contains the following functions:

- `binary_search(term, items)` - The index of the target value in the list, or None if it is not found.
- `simple_search(term, items)` - {List[Result]}: A list of Result objects representing the search results.
- `search_for_least_frequent_items(size, items)` - A list of the specified size containing the least frequent item(s)
- `search_for_most_frequent_items(size, items)` - A list of the specified size containing the most frequent item(s)
"""


from .search import (
  binary_search,
  search_for_least_frequent_items,
  search_for_most_frequent_items,
  simple_search
)



from collections import Counter
import itertools
from typing import Dict, List, Literal, Set, Tuple, Union

from .constants import SUPPORTED_ITEM_TYPES


def default_checker(item_1: SUPPORTED_ITEM_TYPES, item_2: SUPPORTED_ITEM_TYPES) -> Tuple[bool, Union[float, int]]:
    f"""
    Check if two items match, and return a tuple with two values. The first value is a boolean indicating
    whether the items match or not. The second value is a float or an integer, representing a percentage match between 
    the items if they are strings and have a match percentage of more than 50%, or None otherwise.
    
    Args:
    - item_1 ({SUPPORTED_ITEM_TYPES}): An object of a supported type
    - item_2 ({SUPPORTED_ITEM_TYPES}): An object of a supported type
    
    Returns:
    A tuple containing a boolean indicating if the items match, and a float or integer percentage match if the items are
    strings and have a match percentage of more than 50%, or None otherwise.
    """

    if item_1 == item_2:
        return (True, 1)

    if type(item_1) is str and type(item_2) is str and (len(item_1) > 1 or len(item_2) > 1):
        char_counts, combined_chars = char_count([item_1, item_2])

        unconsumed_chars = 0
        total_chars = 0

        for char in combined_chars:
            count_in_item_1 = char_counts[0][char] if char in char_counts[0] else 0
            count_in_item_2 = char_counts[1][char] if char in char_counts[1] else 0

            combined_count = count_in_item_1 + count_in_item_2

            unconsumed_chars += abs(count_in_item_1 - count_in_item_2)
            total_chars += combined_count

        percentage_match = 1 - (unconsumed_chars / max(len(item_1), len(item_2)))
        if percentage_match > 0.5:
            return (True, round(percentage_match, 1))

    return (False, None)


def char_count(items: List[str]) -> Tuple[List[Dict[str, int]], Set[str]]:
    """
    Returns a dictionary mapping each character in a string to the number of times it appears.

    Args:
        s (str): Input string.

    Returns:
        dict: Dictionary mapping each character to the number of times it appears.
    """

    results = []
    combined_chars = set()

    for item in items:
        char_dict = {}
        for char, count in Counter(itertools.chain.from_iterable(item)).items():
            char_dict[char] = count
            combined_chars.add(char)

        results.append(char_dict)

    return (results, combined_chars)


def cleanup_string(s: str) -> str:
    """
    Cleans up a string by removing any leading or trailing white space characters.
    
    Args:
        s (str): The string to be cleaned up.
        
    Returns:
        str: A new string with no leading or trailing white space characters.
    """

    return s.strip()


def heap_sort(size: int, items: List[SUPPORTED_ITEM_TYPES], order: Union[Literal['ASC'], Literal['DESC']]) -> None:
    frequencies_and_indices = {}
    for index, item in enumerate(items):
        data = frequencies_and_indices.get(item, [0, []])
        data[1].append(index)
        data[0] += 1
        frequencies_and_indices[item] = data

    return sorted(
        frequencies_and_indices.items(),
        key=lambda x: x[1][0],
        reverse=(order == 'DESC')
    )[:size]
